/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the code to handle TCP connections to a peer.
 *
 */

#include "waaad-internal.h"
#include "peer-internal.h"

#include <netinet/tcp.h>
#include <netinet/ip6.h>

/* Set the socket options for TCP sockets, before bind is called */
static int _tcp_setsockopt(int family, int sk)
{
	int ret = 0;
	int opt;
	
	/* Clear the NODELAY option in case it was set, as requested by rfc3539#section-3.2 */
	/* Note that this is supposed to be the default, so we could probably remove this call ... */
	opt = 0;
	ret = setsockopt(sk, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
	if (ret != 0) {
		ret = errno;
		TRACE_DEBUG(INFO, "Unable to set the socket TCP_NODELAY option: %s", strerror(ret));
		return ret;
	}
	
	/* Under Linux, we may also set the TCP_CONGESTION option to one of the following strings:
	    - reno (default)
	    - bic
	    - cubic
	    - highspeed
	    - htcp
	    - hybla
	    - illinois
	    - lp
	    - scalable
	    - vegas
	    - veno
	    - westwood
	    - yeah
	*/
	
	/* In case of v6 address, force the v6only option, we use a different socket for v4 */
	#ifdef IPV6_V6ONLY
	if (family == AF_INET6) {
		opt = 1;
		ret = setsockopt(sk, IPPROTO_IPV6, IPV6_V6ONLY, &opt, sizeof(opt));
		if (ret != 0) {
			ret = errno;
			TRACE_DEBUG(INFO, "Unable to set the socket IPV6_V6ONLY option: %s", strerror(ret));
			return ret;
		}
	}
	#endif /* IPV6_V6ONLY */
	
	
	/* There are also others sockopt that can be set, but nothing useful for us AFAICT */
	
	return 0;
}

/*
 * FUNCTION:	_peer_tcp_create_server
 *
 * PARAMETERS:
 *  sock    : (out) On return, the identifier of the bound socket is stored here.
 *  sa      : the socket address on which the server must be bound.
 *  salen   : length of sa structure.
 *
 * DESCRIPTION: 
 *  Creates a listening TCP bound socket.
 *
 * RETURN VALUE:
 *   0 : The socket has been created, bound, and is listening for connections.
 *  !0 : an error occurred (standard errno are returned, such as ENOTSUP)
 */
int _peer_tcp_create_server( int * sock, sSA *addr, socklen_t salen )
{
	TRACE_ENTRY("%p %p %d", sock, addr, salen);
	
	CHECK_PARAMS(  sock  );
	
	/* Create the socket */
	CHECK_SYS(  *sock = socket(addr->sa_family, SOCK_STREAM, IPPROTO_TCP)  );

	/* Set the socket options */
	CHECK_FCT(  _tcp_setsockopt(addr->sa_family, *sock)  );
	
	/* Bind the socket */
	CHECK_SYS(  bind( *sock, addr, salen )  );
		
	/* We're done */
	TRACE_DEBUG(FULL, "TCP socket created: %d", *sock);
	sSA_DUMP( FULL, "TCP socket bound on address: ", addr );
	return 0;
}

/*
 * FUNCTION:	_peer_tcp_connect
 *
 * PARAMETERS:
 *  addr    : primary address of the server, that is passed to connect() function.
 *  addrlen : length of addr structure.
 *  sock    : (out) The client socket is stored here on return.
 *
 * DESCRIPTION: 
 *  Try and connect to a server in TCP.
 *
 * RETURN VALUE:
 *   0 : The client is connected.
 *  !0 : an error occurred (standard errno are returned)
 */
int _peer_tcp_connect( const sSA *addr, socklen_t addrlen, int *sock )
{
	TRACE_ENTRY("%p %d %p", addr, addrlen, sock);
	
	/* Check the parameters */
	CHECK_PARAMS(  addr && addrlen && sock  );
	
	/* Create the socket */
	CHECK_SYS(  *sock = socket(addr->sa_family, SOCK_STREAM, IPPROTO_TCP)  );
	
	/* Set the socket options */
	CHECK_FCT(  _tcp_setsockopt(addr->sa_family, *sock)  );
	
	sSA_DUMP( FULL, "Attempting TCP connection with peer: ", addr );
	
	/* Try connecting to the remote address */
	CHECK_SYS( connect(*sock, addr, addrlen) );
	
	return 0;
}


/* 'Send' and 'receive' functions to pass to the security module */
ssize_t _peer_tcp_send( sec_session_t * sess, void *data, size_t len)
{
	CHECK_PARAMS_DO( sess && sess->conn, { errno = EINVAL; return -1; }  );
	
	return send(*(int *)(sess->conn), data, len, 0);
}

ssize_t _peer_tcp_recv( sec_session_t * sess, void *data, size_t len)
{
	CHECK_PARAMS_DO( sess && sess->conn, { errno = EINVAL; return -1; }  );
	
	return recv(*(int *)(sess->conn), data, len, 0);
}

